unit sockets;

{ By Paul Toth }
{ PaulToth@compuserve.com }
{ http://ourworld.compuserve.com/homepages/paultoth }
{ Tothpaul@Mygale.org }
{ http://www.mygale.org/~tothpaul }

{ This is a beta version of my TCPIP suite }

{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
}interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

{ WINSOCK support }
Const
 WinSock='wsock32.dll'; { 32bits socket DLL }

Type
  TWSAData = packed record
    wVersion: Word;
    wHighVersion: Word;
    szDescription: array[0..256] of Char;
    szSystemStatus: array[0..128] of Char;
    iMaxSockets: Word;
    iMaxUdpDg: Word;
    lpVendorInfo: PChar;
  end;

  TSockAddr=packed record
   Family:word;
   Port:word;
   Addr:LongInt;
   Zero:array[0..7] of byte;
  end;

type
  TFamily=(fmInternet);
  TPort=(poHTTP,poSMTP,poPOP3);
  TEvents=set of (evRead,evWrite,evOOB,evAccept,evConnect,evClose);

  TSocketEvent=procedure(Sender:TObject; ASocket:Integer) of object;
  TErrorEvent=procedure(Sender:TObject; Error:word; ASocket:Integer; Event:word) of object;

Const
 FamilyValue:array[TFamily] of Integer = (2);
 PortValue:array[TPort] of word = (80*256,25*256,110*256);
 scStream=1;

Type
  TSocket = class(TComponent)
  private
   fFamily:TFamily;
   fPort:TPort;
   fIpAddr:string;
   fActive:boolean;
   fEvents:TEvents;
   fServer:boolean;
   EOnRead:TSocketEvent;
   EOnWrite:TSocketEvent;
   EOnConnect:TSocketEvent;
   EOnAccept:TSocketEvent;
   EOnOOB:TSocketEvent;
   EOnClose:TSocketEvent;
   EOnError:TErrorEvent;
   AutoStart:boolean;
   procedure SetActive(Value:boolean);
   procedure SetEvents(Value:TEvents);
   procedure SetServer(Value:boolean);
   procedure WndProc(Var Msg:TMessage);
  protected
   Started:boolean;
   wsaData:TwsaData;
   HSocket:Integer; { should be a word, with $FFFF as invalid socket }
   Handle:THandle;
   fBuffer:Array[0..1024] of char;
   fTextBuf:Array[0..1024] of char;
   fTextPos:word;
   procedure DoRead(ASocket:integer);
   procedure DoWrite(ASocket:integer);
   procedure DoOOB(ASocket:integer);
   procedure DoAccept(ASocket:integer);
   procedure DoConnect(ASocket:integer);
   procedure DoClose(ASocket:integer);
   procedure DoError(ASocket:integer; Error:word; Event:word);
   procedure StartServer;
  public
   constructor Create(AOwner:TComponent); override;
   destructor Destroy; override;
   procedure Loaded; override;
   procedure Call(AHost:string);
   function GetIpAddr(ASocket:Integer):string;
   function GetData(ASocket:integer; Data:PChar; DataLen:word):word;
   function PutData(ASocket:integer; Var Data; DataLen:word):word;
   function WriteLn(ASocket:integer; s:string):word;
   function ReadLn(ASocket:integer; var s:string):integer;
   procedure Close(ASocket:integer);
  published
   Property Family:TFamily read fFamily write fFamily default fmInternet;
   Property Port:TPort read fPort write fPort default poHTTP;
   Property IpAddr:string read fIpAddr write fIpAddr;
   Property Active:boolean read fActive write SetActive;
   Property Events:TEvents read fEvents write SetEvents;
   Property Server:boolean read fServer write SetServer;

   Property OnRead:TSocketEvent read EOnRead write EOnRead;
   Property OnWrite:TSocketEvent read EOnWrite write EOnWrite;
   Property OnAccept:TSocketEvent read EOnAccept write EOnAccept;
   Property OnConnect:TSocketEvent read EOnConnect write EOnConnect;
   Property OnClose:TSocketEvent read EOnClose write EOnClose;
   Property OnOOB:TSocketEvent read EOnOOB write EOnOOB;

   Property OnError:TErrorEvent read EOnError write EOnError;
  end;

procedure Register;

implementation

{ Register }

procedure Register;
begin
  RegisterComponents('MySoft', [TSocket]);
end;

{ Winsock }
Function WSAStartup(Version:word; Var Data:TwsaData):integer; stdcall; far; external winsock;
Function socket(Family,Kind,Protocol:integer):integer; stdcall; far; external winsock;
Function shutdown(Socket,How:Integer):integer; stdcall; far; external winsock;
Function closesocket(socket:Integer):integer; stdcall; far; external winsock;
Function WSACleanup:integer; stdcall; far; external winsock;
Function WSAAsyncSelect(Socket:Integer; Handle:Hwnd; Msg:word; Level:Longint):longint; stdcall; far; external winsock;
Function bind(Socket:Integer; Var SockAddr:TSockAddr; AddrLen:integer):integer; stdcall; far; external winsock;
Function listen(socket,flags:Integer):integer; stdcall; far; external winsock;
Function connect(socket:Integer; Var SockAddr:TSockAddr; AddrLen:integer):integer; stdcall; far; external winsock;
Function accept(socket:Integer; Var SockAddr:TSockAddr; Var AddrLen:Integer):integer; stdcall; far; external winsock;
Function WSAGetLastError:integer; stdcall; far; external winsock;
Function recv(socket:integer; data:pchar; datalen,flags:integer):integer; stdcall; far; external winsock;
Function send(socket:integer; var data; datalen,flags:integer):integer; stdcall; far; external winsock;
Function getpeername(socket:integer; var SockAddr:TSockAddr; Var AddrLen:Integer):Integer; stdcall; far; external winsock;
Function gethostbyname(HostName:PChar):pointer; stdcall; far; external winsock;

Function inet_ntoa(addr:longint):PChar; stdcall; far; external winsock;

Function WSAIsBlocking:boolean; stdcall; far; external winsock;
Function WSACancelBlockingCall:integer; stdcall; far; external winsock;

Function IpToLong(ip:string):LongInt;
 var
  x,i:byte;
  ipx:array[0..3] of byte;
  v:integer;
 begin
  Result:=0;
  longint(ipx):=0; i:=0;
  for x:=1 to length(ip) do
   if ip[x]='.' then begin
    inc(i);
    if i=4 then exit;
   end else begin
    if not (ip[x] in ['0'..'9']) then exit;
    v:=ipx[i]*10+ord(ip[x])-ord('0');
    if v>255 then exit;
    ipx[i]:=v;
   end;
  result:=longint(ipx);
 end;

Function HostToLong(AHost:string):LongInt;
 Type
  PHostEnt = ^THostEnt;
  THostEnt = packed record
    name     : PChar;
    aliases  : ^PChar;
    addrtype : Smallint;
    length   : Smallint;
    addr     : ^pointer;
  end;
 Var
  Host:PHostEnt;
 begin
  Host:=GetHostByName(PChar(AHost));
  if Host=nil then
   Result:=0
  else
   Result:=LongInt(Host^.Addr^^);
 end;

Function LongToIp(Long:LongInt):string;
 var
  ipx:array[0..3] of byte;
  i:byte;
 begin
  longint(ipx):=long;
  Result:='';
  for i:=0 to 3 do result:=result+IntToStr(ipx[i])+'.';
  SetLength(Result,Length(Result)-1);
 end;

{ TSocket }
Constructor TSocket.Create(AOwner:TComponent);
 begin
  inherited Create(AOwner);
  Handle:=AllocateHWnd(WndProc);
 end;

Destructor TSocket.Destroy;
 begin
  Active:=False;
  DeallocateHWnd(Handle);
  inherited Destroy;
 end;

Procedure TSocket.Loaded;
 begin
  inherited Loaded;
  if AutoStart then Active:=True;
 end;

Procedure TSocket.WndProc(Var Msg:TMessage);
 begin
  with msg do
   if msg=wm_user then
    if LParamHi=0 then begin
     case lParamLo of
      $0001 : DoRead(wParam);
      $0002 : DoWrite(wParam);
      $0004 : DoOOB(wParam);
      $0008 : DoAccept(wParam);
      $0010 : DoConnect(wParam);
      $0020 : DoClose(wParam);
     end;
    end else
      DoError(wParam,LParamHi,LParamLo)
   else
    Result:=DefWindowProc(Handle, Msg, wParam, lParam);
 end;

Procedure TSocket.DoRead(ASocket:integer);
 begin
  if Assigned(EOnRead) then EOnRead(Self,ASocket);
 end;

Procedure TSocket.DoWrite(ASocket:integer);
 begin
  if Assigned(EOnWrite) then EOnWrite(Self,ASocket);
 end;

Procedure TSocket.DoOOB(ASocket:integer);
 begin
  if Assigned(EOnOOB) then EOnOOB(Self,ASocket);
 end;

Procedure TSocket.DoAccept(ASocket:integer);
 var
  SockAddr:TSockAddr;
  SockAddrLen:integer;
 begin
  SockAddrLen:=SizeOf(TSockAddr);
  ASocket:=Accept(ASocket,SockAddr,SockAddrLen);
  if Assigned(EOnAccept) then EOnAccept(Self,ASocket);
 end;

Procedure TSocket.DoConnect(ASocket:integer);
 begin
  if Assigned(EOnConnect) then EOnConnect(Self,ASocket);
 end;

Procedure TSocket.DoClose(ASocket:integer);
 begin
  if Assigned(EOnClose) then EOnClose(Self,ASocket);
  ShutDown(ASocket,2);
  CloseSocket(Asocket);
 end;

Procedure TSocket.DoError(ASocket:integer; Error:word; Event:word);
 begin
  if Error=10035 then exit; { WSAEWOULDBLOCK }
  if Assigned(EOnError) then EOnError(Self,Error,ASocket,Event);
 end;

Procedure TSocket.SetActive(Value:boolean);
 begin
  if fActive=value then exit;
  if Value then begin
  { While loading ? }
   if csLoading in ComponentState then begin
    AutoStart:=True;
    exit;
   end;
  { init Winsock DLL }
   Started:=(Started) or (wsaStartup($101,wsaData)=0);
   if not Started then exit;
  { Get socket }
   HSocket:=Socket(FamilyValue[fFamily],scStream,0);
   if HSocket=-1 then exit;
  { allright }
   fActive:=True;
  { Listen on this socket }
   Events:=fEvents;
  { Start Server }
   StartServer;
  end else begin
  { BlockMode }
   if wsaIsBlocking then wsaCancelBlockingCall;
  { free socket }
   if HSocket<>-1 then begin
    ShutDown(HSocket,2);
    CloseSocket(HSocket);
    HSocket:=-1;
   end;
  { done Winsock DLL }
   if Started then begin
    wsaCleanup;
    Started:=False;
   end;
  { done }
   fActive:=False;
  end;
 end;

Procedure TSocket.SetEvents(Value:TEvents);
 begin
  fEvents:=Value;
  if (not Active)or(HSocket=-1) then exit;
  wsaAsyncSelect(HSocket,Handle,wm_user,LongInt(Byte(fEvents)));
 end;

Procedure TSocket.SetServer(Value:boolean);
 begin
  fServer:=Value;
  if (csLoading in ComponentState) then exit;
  if Value then StartServer else begin
  { recreate HSocket }
   Active:=False;
   Active:=True;
  end;
 end;

Procedure TSocket.StartServer;
 var
  SockAddr:TSockAddr;
 begin
  if Active and Server then begin
   SockAddr.Family:=FamilyValue[fFamily];
   SockAddr.Port:=PortValue[fPort];
   SockAddr.Addr:=IpToLong(fIpAddr);
   if Bind(HSocket,SockAddr,SizeOf(SockAddr))<>0 then
    DoError(HSocket,wsaGetLastError,0)
   else
    if Listen(HSocket,0)<>0 then DoError(HSocket,wsaGetLastError,0);
  end;
 end;

Procedure TSocket.Call(AHost:String);
 Var
  SockAddr:TSockAddr;
 begin
  fTextBuf[0]:=#0;
  SockAddr.Family:=FamilyValue[fFamily];
  SockAddr.Port:=PortValue[fPort];
  SockAddr.Addr:=IpToLong(AHost);
  if SockAddr.Addr=0 then begin
   SockAddr.Addr:=HostToLong(AHost);
  end;
  if Connect(HSocket,SockAddr,SizeOf(SockAddr))<>0 then DoError(HSocket,wsaGetLastError,0);
 end;

Function TSocket.GetIpAddr(ASocket:Integer):string;
 Var
  SockAddr:TSockAddr;
  SockLen:Integer;
 begin
  SockLen:=SizeOf(SockAddr);
  if GetPeerName(ASocket,SockAddr,SockLen)=0 then
   Result:=LongToIp(SockAddr.Addr)
  else
   Result:='???.???.???.???';
 end;

Function TSocket.GetData(ASocket:integer; Data:PChar; DataLen:word):word;
 begin
  Result:=recv(ASocket,Data,DataLen,0);
  if Result=$FFFF then begin
   DoError(ASocket,wsaGetLastError,word(evRead));
   Result:=0;
  end;
 end;

Function TSocket.PutData(ASocket:integer; Var Data; DataLen:word):word;
 begin
  Move(Data,fBuffer,DataLen);
  Result:=send(ASocket,fBuffer,DataLen,0);
  if Result=$FFFF then begin
   DoError(ASocket,wsaGetLastError,word(evWrite));
   Result:=0;
  end;
 end;

Function TSocket.WriteLn(ASocket:integer; s:string):word;
 begin
  s:=s+#13#10;
  Result:=PutData(ASocket,PChar(s)^,Length(s));
 end;

Function TSocket.readLn(ASocket:integer; var s:string):integer;
 var
  CR,LF:PChar;
 begin
  Result:=-1;
  fTextPos:=StrLen(fTextBuf);
  Result:=GetData(ASocket,@fTextBuf[fTextPos],SizeOf(fTextBuf)-fTextPos);
  inc(fTextPos,Result);
  if fTextPos>0 then begin
   fTextBuf[fTextPos]:=#0;

   CR:=StrPos(fTextBuf,#13);
   if CR=nil then
    CR:=StrPos(fTextbuf,#10)
   else begin
    CR^:=#0;
    LF:=StrPos(fTextBuf,#10);
    if LF<>nil then begin CR^:=#13; CR:=LF; end;
   end;

   if CR=nil then
    Result:=-1
   else begin
    CR^:=#0;
    s:=fTextBuf;
    inc(CR); if CR^ in [#10,#13] then inc(CR);
    StrCopy(fTextBuf,CR);
    Result:=Length(s);
   end;
  end else
   Result:=-1;
 end;

Procedure TSocket.close(ASocket:integer);
 begin
  DoClose(ASocket);
 end;

end.
